from .client import init
from .frameworks import flask_rosnik
from .events.user import track_user_feedback, track_user_interaction

__all__ = ["init", "flask_rosnik", "track_user_interaction", "track_user_feedback"]



import functools
import logging
import time
from typing import Callable, Iterator, Union

from rosnik import constants
from rosnik.events import queue
from rosnik.types.core import AIEvent, Metadata
from rosnik.wrap import wrap_class_method
from rosnik.types.ai import (
    AIFunctionMetadata,
    AIRequestFinish,
    AIRequestStartStream,
    AIRequestStart,
    ErrorResponseData,
    OpenAIAttributes,
)

try:
    from openai.openai_object import OpenAIObject
except ImportError as e:
    raise e

logger = logging.getLogger(__name__)

_OAI = "openai"


def hook_with_metadata(hook: Callable, generate_metadata: Callable[[], AIFunctionMetadata]):
    """Each provider should be encapsulated into what it knows about the call."""
    return functools.partial(hook, generate_metadata=generate_metadata)


def request_hook(
    payload: dict,
    function_fingerprint: str,
    prior_event: AIEvent = None,
    generate_metadata: Callable[[], AIFunctionMetadata] = None,
) -> AIRequestStart:
    """`payload` is a dictionary of the `kwargs` provided to `create`.

    Given those kwargs and metadata, generate a AIRequestStart event.
    """
    if not payload:
        return None

    if generate_metadata is None:
        logger.warning("`generate_metadata` not provided. Bailing.")
        return

    # Reimport to make sure we have the latest values
    import openai

    metadata = generate_metadata()
    metadata.openai_attributes = OpenAIAttributes(
        api_base=openai.api_base,
        api_type=openai.api_type,
        api_version=openai.api_version,
        organization=openai.organization,
    )

    ai_model = (
        payload.get("model")
        if openai.api_type == "open_ai"
        else (payload.get("deployment_id") or payload.get("engine"))
    )
    user_id = payload.get("user")

    event = AIRequestStart(
        ai_model=ai_model,
        ai_provider=metadata.ai_provider,
        ai_action=metadata.ai_action,
        ai_metadata=metadata,
        request_payload=payload,
        user_id=user_id,
        _metadata=Metadata(
            function_fingerprint=function_fingerprint, stream=payload.get("stream", False)
        ),
    )
    queue.enqueue_event(event)
    return event


def response_hook(
    payload: Union["OpenAIObject", Iterator],
    function_fingerprint: str,
    prior_event: AIEvent = None,
    generate_metadata: Callable[[], AIFunctionMetadata] = None,
) -> AIRequestFinish:
    """If we're an Iterator, it means we're a streamed response,
    in which case we're tracking first-byte here.
    If we're an OpenAIObject, then we can provide additional metadata.
    """
    if not payload:
        return None

    if generate_metadata is None:
        logger.warning("`generate_metadata` not provided. Bailing.")
        return

    # Reimport to make sure we have the latest values
    import openai

    is_stream_response = isinstance(payload, Iterator)

    metadata = generate_metadata()
    metadata.openai_attributes = OpenAIAttributes(
        api_base=openai.api_base,
        api_type=openai.api_type,
        api_version=openai.api_version,
        organization=openai.organization,
    )

    now = int(time.time_ns() / 1000000)
    event_kwargs = {
        # For both OpenAI and Azure, `model` contains the model on the response.
        "ai_model": prior_event.ai_model if is_stream_response else payload.get("model"),
        "ai_provider": metadata.ai_provider,
        "ai_action": metadata.ai_action,
        "ai_metadata": metadata,
        "response_payload": payload,
        "sent_at": now,
        "response_ms": (now - prior_event.sent_at),
        "ai_request_start_event_id": prior_event.event_id,
        "user_id": prior_event.user_id,
        "_metadata": Metadata(function_fingerprint=function_fingerprint, stream=is_stream_response),
    }

    if is_stream_response:
        # These don't exist. `payload` is a generator.
        # Make this None
        event_kwargs.pop("response_payload")
        event = AIRequestStartStream(**event_kwargs)
        queue.enqueue_event(event)
        return event

    event = AIRequestFinish(**event_kwargs)
    queue.enqueue_event(event)
    return event


def streamed_response_hook(
    response: Iterator,
    function_fingerprint: str,
    prior_event: AIEvent = None,
    generate_metadata: Callable[[], AIFunctionMetadata] = None,
):
    """Wrap the response generator with our own function so that the
    user can still yield results, and we can automatically
    figure out the duration of the stream.
    """

    content_parts = []
    # Load openai here and use values defined at this point.
    import openai

    def _stream_hook(line: "OpenAIObject"):
        """For each chunk, either add it to our content_parts
        or emit an event because we're done.
        """
        if not line:
            logger.debug("Line not seen on stream.")
            return

        choices = line.get("choices")
        if not choices or len(choices) < 1:
            logger.debug("Missing choices on stream.")
            return

        content = choices[0].get("delta", {}).get("content")
        if content:
            content_parts.append(content)

        finish_reason = choices[0].get("finish_reason")
        if finish_reason:
            metadata = generate_metadata()
            metadata.openai_attributes = OpenAIAttributes(
                api_base=openai.api_base,
                api_type=openai.api_type,
                api_version=openai.api_version,
                organization=openai.organization,
            )
            now = int(time.time_ns() / 1000000)
            finish_event = AIRequestFinish(
                ai_model=line.get("model"),
                ai_provider=metadata.ai_provider,
                ai_action=metadata.ai_action,
                ai_metadata=metadata,
                # Mimic the OpenAI response payload.
                # Missing `usage`, TODO: what do?
                response_payload={
                    "choices": [
                        {
                            "finish_reason": finish_reason,
                            "index": 0,
                            "message": {"content": "".join(content_parts), "role": "assistant"},
                        }
                    ],
                    "model": line.get("model"),
                    "created": line.get("created"),
                    "object": line.get("object"),
                    "id": line.get("id"),
                },
                sent_at=now,
                response_ms=(now - prior_event.sent_at),
                ai_request_start_event_id=prior_event.event_id,
                user_id=prior_event.user_id,
                _metadata=Metadata(function_fingerprint=function_fingerprint, stream=True),
            )
            queue.enqueue_event(finish_event)

    def _stream_response_wrapper(response: Iterator):
        for line in response:
            _stream_hook(line)
            yield line

    return _stream_response_wrapper(response)


def error_hook(
    error: Exception,
    function_fingerprint: str,
    prior_event: AIEvent = None,
    generate_metadata: Callable[[], AIFunctionMetadata] = None,
) -> AIRequestFinish:
    if error is None:
        return None

    if generate_metadata is None:
        logger.warning("`generate_metadata` not provided. Bailing.")
        return

    import openai

    metadata = generate_metadata()
    metadata.openai_attributes = OpenAIAttributes(
        api_base=openai.api_base,
        api_type=openai.api_type,
        api_version=openai.api_version,
        organization=openai.organization,
    )

    now = int(time.time_ns() / 1000000)
    event = AIRequestFinish(
        ai_model=None,
        ai_provider=metadata.ai_provider,
        ai_action=metadata.ai_action,
        ai_metadata=metadata,
        error_data=ErrorResponseData(
            message=str(error),
            json_body=getattr(error, "json_body", None),
            headers=getattr(error, "headers", None),
            organization=getattr(error, "organization", None),
            request_id=getattr(error, "request_id", None),
        ),
        response_ms=(now - prior_event.sent_at),
        ai_request_start_event_id=prior_event.event_id,
        user_id=prior_event.user_id,
        _metadata=Metadata(function_fingerprint=function_fingerprint),
    )
    queue.enqueue_event(event)
    return event


def _patch_completion(openai):
    if getattr(openai, f"__{constants.NAMESPACE}_patch", False):
        logger.warning("Not patching. Already patched.")
        return

    wrap_class_method(
        openai.Completion,
        "create",
        hook_with_metadata(
            request_hook,
            lambda: AIFunctionMetadata(ai_provider=_OAI, ai_action="completions"),
        ),
        hook_with_metadata(
            response_hook,
            lambda: AIFunctionMetadata(ai_provider=_OAI, ai_action="completions"),
        ),
        hook_with_metadata(
            error_hook, lambda: AIFunctionMetadata(ai_provider=_OAI, ai_action="completions")
        ),
        hook_with_metadata(
            streamed_response_hook,
            lambda: AIFunctionMetadata(ai_provider=_OAI, ai_action="completions"),
        ),
    )


def _patch_chat_completion(openai):
    if getattr(openai, f"__{constants.NAMESPACE}_patch", False):
        logger.warning("Not patching. Already patched.")
        return

    wrap_class_method(
        openai.ChatCompletion,
        "create",
        hook_with_metadata(
            request_hook,
            lambda: AIFunctionMetadata(ai_provider=_OAI, ai_action="chat.completions"),
        ),
        hook_with_metadata(
            response_hook,
            lambda: AIFunctionMetadata(ai_provider=_OAI, ai_action="chat.completions"),
        ),
        hook_with_metadata(
            error_hook,
            lambda: AIFunctionMetadata(ai_provider=_OAI, ai_action="chat.completions"),
        ),
        hook_with_metadata(
            streamed_response_hook,
            lambda: AIFunctionMetadata(ai_provider=_OAI, ai_action="chat.completions"),
        ),
    )


def _patch_openai():
    try:
        import openai
    except ImportError as e:
        raise e

    if getattr(openai, f"__{constants.NAMESPACE}_patch", False):
        logger.warning("Not patching. Already patched.")
        return

    if getattr(openai, "Completion", None):
        _patch_completion(openai)
    if getattr(openai, "ChatCompletion", None):
        _patch_chat_completion(openai)
    setattr(openai, f"__{constants.NAMESPACE}_patch", True)



import os
from rosnik import constants

API_KEY = f"{constants.NAMESPACE}_API_KEY"
# If set to a non-0 value,
# we will send values synchronously.
SYNC_MODE = f"{constants.NAMESPACE}_SYNC_MODE"
ENVIRONMENT = f"{constants.NAMESPACE}_ENVIRONMENT"


class _Config:
    def __init__(self, api_key=None, sync_mode=None, environment=None):
        self._api_key = api_key or os.environ.get(API_KEY)
        _sync = sync_mode or os.environ.get(SYNC_MODE)
        self._sync_mode = _sync and _sync != "0"
        self._environment = environment or os.environ.get(ENVIRONMENT)

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, value):
        if self._api_key is not None:
            return
        self._api_key = value

    @property
    def sync_mode(self):
        return bool(self._sync_mode)

    @sync_mode.setter
    def sync_mode(self, value):
        if self._sync_mode is not None:
            return
        self._sync_mode = value

    @property
    def environment(self):
        return self._environment

    @environment.setter
    def environment(self, value):
        if self._environment is not None:
            return
        self._environment = value


Config = _Config()



import logging
from urllib3.util import Retry

from rosnik import config
from rosnik.types import core

import requests
from requests.adapters import HTTPAdapter

logger = logging.getLogger(__name__)

_base_url = "https://ingest.rosnik.ai/api/v1/events"

_NUM_RETRIES = 3

_retry_status_code = [429, 500, 502, 503, 504]
retry_strategy = Retry(
    total=_NUM_RETRIES,
    backoff_factor=1,
    status_forcelist=_retry_status_code,
    allowed_methods=["POST"],
)
adapter = HTTPAdapter(max_retries=retry_strategy)


class IngestClient:
    def __init__(self):
        self.session = requests.Session()
        self.session.mount("https://", adapter)

    @property
    def headers(self):
        return {
            "Authorization": f"Bearer {config.Config.api_key}",
            "Content-Type": "application/json",
        }

    def _post(self, *args, **kwargs):
        # Wait up to 3 seconds before giving up
        kwargs["timeout"] = 3
        return self.session.post(*args, **kwargs)

    def send_event(self, event: core.Event, url=_base_url):
        logger.debug(
            f"Sending {event.event_type} event to {url} with event ID {event.event_id} and journey ID {event.journey_id}"  # noqa
        )
        try:
            response = self._post(url, headers=self.headers, json=event.to_dict())
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            logger.warning(f"Failed to send event: {e}")
        except requests.exceptions.RetryError as e:
            logger.warning(f"Failed to send event after {_NUM_RETRIES} attempts: {e}")



from dataclasses import dataclass
from typing import Optional
from rosnik.types.core import UserEvent


@dataclass(kw_only=True, slots=True)
class UserGoalSuccess(UserEvent):
    event_type: str = "user.goal.success"
    goal_name: str


@dataclass(kw_only=True, slots=True)
class UserInteractionTrack(UserEvent):
    event_type: str = "user.interaction.track"
    # One of: ai-request, text-edit
    interaction_type: str


@dataclass(kw_only=True, slots=True)
class UserFeedbackTrack(UserEvent):
    """Explicit feedback"""

    event_type: str = "user.feedback.track"
    score: Optional[int]
    open_response: Optional[str]



"""
A request on its own can be considered a journey,
if there is not an external journey ID supplied.

We manage state here across the 3 values we care about.
Outside of Journey ID, they can be set to None, which
would mean they are user-less journeys.
"""
from enum import Enum
import logging
from contextvars import ContextVar
from typing import Optional

import ulid

logger = logging.getLogger(__name__)

_journey_id_key = "journey_id"
_user_interaction_id_key = "user_interaction_id"
_device_id_key = "device_id"

journey_id_cv: ContextVar[Optional[str]] = ContextVar(_journey_id_key)
user_interaction_id_cv: ContextVar[Optional[str]] = ContextVar(_user_interaction_id_key)
device_id_cv: ContextVar[Optional[str]] = ContextVar(_device_id_key)


class State(Enum):
    JOURNEY_ID = _journey_id_key
    USER_INTERACTION_ID = _user_interaction_id_key
    DEVICE_ID = _device_id_key


_state = {
    State.JOURNEY_ID: journey_id_cv,
    State.USER_INTERACTION_ID: user_interaction_id_cv,
    State.DEVICE_ID: device_id_cv,
}


def store(state_type: State, value: str):
    _state[state_type].set(value)


def retrieve(state_type: State):
    return _state[state_type].get(None)


def create_journey_id():
    journey_id = ulid.new().str
    store(State.JOURNEY_ID, journey_id)
    return journey_id


def get_journey_id():
    journey_id = retrieve(State.JOURNEY_ID)
    if journey_id is None:
        return create_journey_id()
    return journey_id


def get_device_id():
    return retrieve(State.DEVICE_ID)


def get_user_interaction_id():
    return retrieve(State.USER_INTERACTION_ID)


def _reset():
    journey_id_cv.set(None)
    user_interaction_id_cv.set(None)
    device_id_cv.set(None)



import time
import os
import platform
from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import DataClassJsonMixin
import ulid

from rosnik import state


def _generate_event_id():
    return ulid.new().str


@dataclass(kw_only=True, slots=True)
class StaticMetadata:
    environment: Optional[str] = os.environ.get("ROSNIK_ENVIRONMENT", None)
    runtime: str = platform.python_implementation()
    runtime_version: str = platform.python_version()
    # TODO: how to sync pyproject version to this
    sdk_version: str = "0.0.27"


@dataclass(kw_only=True, slots=True)
class Metadata(StaticMetadata):
    function_fingerprint: str
    stream: bool = False


@dataclass(kw_only=True, slots=True)
class Event(DataClassJsonMixin):
    # Unique event ID
    event_id: str = field(default_factory=_generate_event_id)
    event_type: str
    journey_id: str = field(default_factory=state.get_journey_id)
    # Epoch in ms
    sent_at: int = int(time.time_ns() / 1000000)
    # Epoch unless not set, which will be -1
    occurred_at: Optional[int] = None
    # JSONable user defined context
    # TODO: TBD on how this is provided.
    context: Optional[dict] = None
    # Users could be part of an AI event or a User event
    user_id: Optional[str] = None
    device_id: Optional[str] = field(default_factory=state.get_device_id)
    # Our own metadata
    _metadata: Metadata
    # User Interaction ID: this is the causal user.interaction.track
    # event ID for this event. If it's unset then something else
    # we're not tracking caused this action.
    user_interaction_id: Optional[str] = field(default_factory=state.get_user_interaction_id)


@dataclass(kw_only=True, slots=True)
class AIEvent(Event):
    # model name
    ai_model: str
    # openai
    ai_provider: str
    # AI action: completion, chatcompletion
    ai_action: str


class UserEvent(Event):
    # JSONable user properties that can be added
    user_properties: dict
